/* Copyright (C) 1989, 1995, 1996 Aladdin Enterprises.  All rights reserved.
  
  This file is part of Aladdin Ghostscript.
  
  Aladdin Ghostscript is distributed with NO WARRANTY OF ANY KIND.  No author
  or distributor accepts any responsibility for the consequences of using it,
  or for whether it serves any particular purpose or works at all, unless he
  or she says so in writing.  Refer to the Aladdin Ghostscript Free Public
  License (the "License") for full details.
  
  Every copy of Aladdin Ghostscript must include a copy of the License,
  normally in a plain ASCII text file named PUBLIC.  The License grants you
  the right to copy, modify and redistribute Aladdin Ghostscript, but only
  under certain conditions described in the License.  Among other things, the
  License requires that the copyright notice and this notice be preserved on
  all copies.
*/

/* gdevepag.c   gdevepag Ver.3*/  
/* $Id: gdevepag.c,v 1.2 1998/03/31 08:40:09 nari Exp $ */
/* Many parts of this file are copied from gdevdjet.c */
/* Epson ESC/Page driver for Ghostscript */

#include "gdevprn.h"

#define ESC 0x1b    /* \033 */
#define GS  0x1d    /* \035 */

/* cut from gdevpcl.h gdevmjc.c */
#define PAPER_SIZE_LETTER 2
#define PAPER_SIZE_LEGAL 3
#define PAPER_SIZE_A4 26
#define PAPER_SIZE_A3 27
#define PAPER_SIZE_A3_NOBI 28
#define PAPER_SIZE_A2 29
#define PAPER_SIZE_B4 30
#define PAPER_SIZE_B5 31

/* The device descriptors */
private dev_proc_open_device(epag_open);
private dev_proc_close_device(epag_close);
private dev_proc_print_page(epag_print_page);
private dev_proc_print_page_copies(epag_print_copies);
private void epag_printer_initialize(gx_device_printer *pdev, FILE *fp,int);
private dev_proc_print_page(epag_print_page);
private int epag_flush(gx_device_printer *pdev, FILE *fp, int bpl, int cRow,
			byte *ibuf);
private int gdev_epag_paper_size(gx_device_printer *dev);
private dev_proc_get_params(epag_get_params);
private dev_proc_put_params(epag_put_params);

#define X_DPI 300
#define Y_DPI 300

typedef struct gx_device_epag_printer_s {
  gx_device_common;
  gx_prn_device_common;
  int bBoth;
  int bFlip;
  float offX;   /* Xեåȡ ¦ ñ mm */
  float offY;   /* Yեåȡ ¦ ñ mm */
  int cRowBuf;
} gx_device_epag_printer; 

#define epag ((gx_device_epag_printer *) pdev)
private gx_device_procs prn_epag_procs =
prn_params_procs(epag_open, gdev_prn_output_page, epag_close,
		 epag_get_params, epag_put_params);

gx_device_printer far_data gs_epag_device =
prn_device(prn_epag_procs, "epag", 
		DEFAULT_WIDTH_10THS, DEFAULT_HEIGHT_10THS,
		X_DPI, Y_DPI,
		0.0, 0.0, 0.0, 0.0,	/* margins */
		1, epag_print_page);

#define ppdev ((gx_device_printer *)pdev)

/* Open the printer, adjusting the margins if necessary. */
private int
epag_open(gx_device *pdev)
{
  /* Change the margins if necessary. */
  const float _ds *m = 0;

  if ( m != 0 )
    gx_device_set_margins(pdev, m, true);
	
  /* enable Duplex. */
  ppdev->Duplex = true, ppdev->Duplex_set = 0;
	
  return gdev_prn_open(pdev);
}

private int
epag_close(gx_device *pdev)
{
  gdev_prn_open_printer(pdev, 1);
  if((epag->bFlip || epag->bBoth) && (pdev->PageCount & 1)){
    fprintf(ppdev->file ,"%c%dY%c%dX", GS, 0, GS, 0);
    fprintf(ppdev->file,"Blank Page\r\n");
    fprintf(ppdev->file,"%c0dpsE",GS);
  }
#if 0
  fprintf(ppdev->file,"%crhE",GS);
#endif
  return gdev_prn_close(pdev);
}

#undef ppdev

private int
epag_print_page(gx_device_printer *pdev, FILE *fp)
{
  int code = 0;
  int cRowBuf;         /* ٤˽Ϥ rowο */
  int cRow;            /* bufäƤrowο */
  int bpl;             /* byte per line */
  byte *ibuf,*bp;      /* image buffer  */
  int x,y,ri;
  
  /* alloc buffers */

  if(epag->cRowBuf < 1)
    epag->cRowBuf = 128;
  cRowBuf = epag->cRowBuf;
  
  bpl = gdev_mem_bytes_per_scan_line((gx_device *)pdev);
  ibuf = (byte *)gs_malloc(bpl * cRowBuf, 1, "epag_print_page(ibuf)");
  if(ibuf == NULL) return_error(gs_error_VMerror);

  /* printer initialize */
  
  if(pdev->PageCount == 0)
    epag_printer_initialize(pdev, fp, 1);

  /* image output */
  
  bp = ibuf;
  cRow = 0;
  x = (epag->offX - 3.0) / 25.4 * pdev->x_pixels_per_inch; 
  y = (epag->offY + 3.0) / 25.4 * pdev->y_pixels_per_inch; 
  fprintf(fp ,"%c%dY%c%dX", GS, y, GS, x);
  for(ri = 0; ri < pdev->height; ri++){
    code = gdev_prn_copy_scan_lines(pdev, ri, bp, bpl);
    if(code < 0) return code;
    bp += bpl;
    if(++cRow >= cRowBuf){
      code = epag_flush(pdev, fp, bpl, cRow, ibuf);
      if(code < 0) return code;
      cRow = 0;
      bp = ibuf;
    }
  }
  if(cRow > 0){
    code = epag_flush(pdev, fp, bpl, cRow, ibuf);
    if(code < 0) return code;
  }

  if(epag->bBoth || epag->bFlip)
    fprintf(fp,"%c0dpsE",GS);
  else
    fprintf(fp,"\014"); /* eject page */
  gs_free(ibuf, cRowBuf * bpl, 1, "epag_print_page(ibuf)");
  return code;
}


/* output printer initialize code */

/* ------ Internal routines ------ */

static char can_inits[] ={
  ESC, 'S', ESC, ESC, 'S', 034,
  ESC, 'z', 0, 0,                  /* ESC Page ⡼ɤ  */
  GS, 'r', 'h', 'E',               /* hard reset         */
  GS, '0', 'p', 'o', 'E',          /* Portrait           */
  GS, '1', 'm', 'm', 'E',          /* ڡ⡼ */
  GS, '3', 'b', 'c', 'I',          /* ̷           */
  GS, '2', 'a', 'b', 'P',          /* ᡼岼 */
};


private
void epag_printer_initialize(gx_device_printer *pdev, FILE *fp, int copies)
{
  double xDpi,yDpi;
  char *pse = NULL;
  int paper_size;
  
  xDpi = pdev->x_pixels_per_inch;
  yDpi = pdev->y_pixels_per_inch;

  /*  */
  
  fwrite(can_inits, sizeof(can_inits), 1, fp);
  
  /* ξ̰ */

  if(epag->bBoth){
    fprintf(fp,"%c1sdE",GS);
    fprintf(fp,"%c0bdE",GS);
  }
  else if(epag->bFlip){
    fprintf(fp,"%c1sdE",GS);
    fprintf(fp,"%c1bdE",GS);
  }
  else
    fprintf(fp,"%c0sdE",GS);
    
  /* ѻ極 */
  
  paper_size=gdev_epag_paper_size(pdev);
  switch(paper_size){
 case PAPER_SIZE_LETTER:  pse="14psE";break;
 case PAPER_SIZE_LEGAL:   pse="14psE";break;
 case PAPER_SIZE_A4:      pse="14psE";break;
 case PAPER_SIZE_A3:      pse="13psE";break;
 case PAPER_SIZE_A3_NOBI: pse="13psE";break;
 case PAPER_SIZE_A2:      pse="12psE";break;
 case PAPER_SIZE_B4:      pse="24psE";break;
 case PAPER_SIZE_B5:      pse="25psE";break;
       }

  if(pse) fprintf(fp,"%c%s",GS,pse);
  
  /* Ǿñ  */
  
  fprintf(fp,"%c0;%4.2fmuE",GS, 72.0 / xDpi );

  /*      */
  
  fprintf(fp,"%c0;%d;%ddrE",GS, (int)(xDpi+0.5),(int)(yDpi+0.5));

  /* ԡ */

  if(copies > 1)
    fprintf(fp,"%c%dcoO",GS, copies);
}

private
int epag_flush(gx_device_printer *pdev, FILE *fp, int bpl, int cRow,
		byte *ibuf)
{
  byte *cbuf,*cp,*bp,*bp1,*bp2,*bpMax;
  int  size,len;

  size = bpl * cRow / 2 * 3; 
  cbuf = (byte *)gs_malloc(size, 1, "epag_flush(cbuf)");
  if(cbuf == NULL) return_error(gs_error_VMerror);

  bp = ibuf;
  bpMax = ibuf + bpl * cRow;
  cp = cbuf;
  while(bp < bpMax){
    bp2 = bp + 256;
    if(bp2 > bpMax) bp2 = bpMax;
    for(bp1 = bp+1; (bp1 < bp2) && bp[0] == bp1[0]; bp1++);
    len = bp1 - bp;
    *cp++ = bp[0];
    if(len > 1) {
      *cp++ = bp[0];
      *cp++ = len - 2;
    }
    bp = bp1;
  }
  fprintf(fp,"%c%d;%d;%d;0bi{I",GS,cp-cbuf,bpl*8,cRow);
  fwrite(cbuf, cp-cbuf, 1, fp);
  gs_free(cbuf, size, 1, "epag_flush(cbuf)");
}

private int
gdev_epag_paper_size(gx_device_printer *dev)
{ return
      (dev->height / dev->y_pixels_per_inch >= 22.2 ? PAPER_SIZE_A2 :
       dev->height / dev->y_pixels_per_inch >= 18.0 ? PAPER_SIZE_A3_NOBI :
       dev->height / dev->y_pixels_per_inch >= 16.0 ? PAPER_SIZE_A3 :
       dev->height / dev->y_pixels_per_inch >= 13.6 ? PAPER_SIZE_B4 :
       dev->height / dev->y_pixels_per_inch >= 11.8 ? PAPER_SIZE_LEGAL :
       dev->height / dev->y_pixels_per_inch >= 11.1 ? PAPER_SIZE_A4 :
       dev->height / dev->y_pixels_per_inch >= 10.4 ? PAPER_SIZE_LETTER :
       PAPER_SIZE_LETTER);
}

private int
epag_get_params(gx_device *pdev, gs_param_list *plist)
{
  gdev_prn_get_params(pdev, plist);
  param_write_int(plist, "cRowBuf", &epag->cRowBuf);
  param_write_int(plist, "Flip", &epag->bFlip);
  param_write_int(plist, "Both", &epag->bBoth);
  param_write_float(plist, "EpagOffX", &epag->offX);
  param_write_float(plist, "EpagOffY", &epag->offY);
  return 0;
}

/* Put properties. */
private int
epag_put_params(gx_device *pdev, gs_param_list *plist)
{
  param_read_int(plist, "cRowBuf", &epag->cRowBuf);
  param_read_int(plist, "Flip", &epag->bFlip);
  param_read_int(plist, "Both", &epag->bBoth);
  param_read_float(plist, "EpagOffX", &epag->offX);
  param_read_float(plist, "EpagOffY", &epag->offY);
  gdev_prn_put_params(pdev, plist);
  return 0;
}
